;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "autothemer" "0.2.18"
  "Conveniently define themes."
  '((dash  "2.10.0")
    (emacs "26.1"))
  :url "https://github.com/jasonm23/autothemer"
  :commit "0d3ed8d74cdf9be90492999217db1bc7ba866607"
  :revdesc "0d3ed8d74cdf"
  :maintainers '(("Jason Milkins" . "jasonm23@gmail.com")))
